#!/usr/bin/env python3
#
# LNK ReShuffle - tool to extract/replace DOA5LR assets from LNK archives
#
# version: 1  (20150512)
# author: immi
#
# Feel free to distribute, modify, adapt or reuse the code
#
# Credits to the authors of doa5LRarchivarius and all the other people who
# did the hard work of figuring out the file format.

__appname__='LNK_ReShuffle'
__version__='v1'

import io
import os
import struct
from sys import argv
from time import time

from tkinter import *
from tkinter import ttk
from tkinter import filedialog

import zlib

DAT_DIR=os.path.join(os.path.dirname(argv[0]),'dat')

# yeah TeamNinja, we hear you :p
XOR_PHRASE = b"Except as expressly authorized, it is strictly prohibited to reproduce, distribute, exhibit or modify this software and any of its contents, including audio and visual contents. By way of example, to capture, copy or download any of the contents in this software, including audio and visual contents, onto any hardware or other software source media for any purpose, by the Internet or any other source, is strictly prohibited. Reversed engineering, decompiling or disassembly of this software is also strictly prohibited."

###############
# helper

class Error(Exception):
        # blaa
        pass

def get_zstr(buf, ofs):
        idx = buf.find(b'\x00', ofs)
        return buf[ofs:idx].decode('ASCII')

def buf_str(buf):
        return "".join(["%02x"%i for i in buf])

def align(num, boundary):
        x = boundary-1
        return  (num + x) & ~x

class AssetInfo:
        def __init__(self,name,lnk_name,type, unk,idx):
                self.name=name
                self.alias=name
                self.lnk=lnk_name
                self.type=type
                self.unk=unk
                self.idx=idx
                self.size=0
                self.flags=0
                #self.size2=0
                self.crc=0
                self.lnk_idx=None
                self.lnk_size=None
                self.lnk_ofs=None
                self.lnk_name=None


##############################
# handle game.exe

# table are directly next to each other
ASSET_TABLE_OFS=0xbe98e0
CRC_TABLE_OFS=0xc3aa50
INFO_TABLE_OFS=0xc4d268
RVA_DIFF=0x401800 # VA -> file offset

def load_exe_strings(fd, ofs_list):
        start = min(ofs_list) - RVA_DIFF
        end = max(ofs_list) - RVA_DIFF

        fd.seek(start)
        buf=fd.read(end-start+128) # assume the last string < 128chars
        for i in range(len(ofs_list)):
                ofs_list[i] = get_zstr(buf,ofs_list[i]-start-RVA_DIFF)

def load_exe_info(exe_path):
        asset_map={}
        with open(exe_path,'rb') as fs:
                fs.seek(ASSET_TABLE_OFS)
                buf=fs.read(CRC_TABLE_OFS-ASSET_TABLE_OFS)
                num=len(buf)//16 # 4dwords
                asset_table = struct.unpack('<%dL'%(num*4),buf)

                # 1th,3rd element are string offsets
                strings=list(asset_table[::4])
                strings+=list(asset_table[2::4])
                load_exe_strings(fs, strings)

                fs.seek(INFO_TABLE_OFS)
                buf=fs.read(num*8) # same number of entries as asset table
                info_table=struct.unpack_from("<%dL"%(num*2),buf)

                fs.seek(CRC_TABLE_OFS)
                buf=fs.read(INFO_TABLE_OFS-CRC_TABLE_OFS)
                crc_table=struct.unpack("<%dL"%(len(buf)/4),buf)

        for i in range(num):
                asset=AssetInfo(strings[i],strings[i+num],asset_table[i*4+1],asset_table[i*4+3],i)
                asset.size = info_table[i*2]
                asset.flags= info_table[i*2+1]
                if asset.flags&0x80000000:#FIXME re-check
                        idx = asset.flags&0xffff
                        asset.size=crc_table[idx*2] # other size seems irrelevant??
                        asset.crc=crc_table[idx*2+1]

                asset_map[asset.name]=asset
        print(" ..read %d asset entries from game.exe"%len(asset_map))
        return asset_map

def fixup_exe(exe_path, asset, new_size):
        if asset.size == new_size: return

        if asset.flags&0x80000000: # FIXME check this
                ofs=CRC_TABLE_OFS
                idx=asset.flags&0xffff
        else:
                ofs=INFO_TABLE_OFS
                idx=asset.idx
        with open(exe_path, 'r+b') as fd:
                fd.seek(ofs+idx*8)
                fd.write( struct.pack('<L',new_size) )

                if asset.flags&0x80000000: # FIXME re-check this logic
                        ofs=INFO_TABLE_OFS
                        idx=asset.idx
                        fd.seek(ofs+idx*8+4)
                        # remove (1<<29) flag -> disable crc check
                        fd.write( struct.pack('<L',asset.flags& ~0x20000000) )
                        asset.flags = asset.flags& ~0x20000000
                        print(" ..fixed flags to:%#08x"%asset.flags)

        print(" ..fixed up game.exe @%#08x -> size:%#06x"%(ofs,new_size))

############################
# LNK stuff

# FIXME: propably can remove this
FTYPE={'texc','dlic','char_','PHYD'}
def guess_is_compressed(buf):
        try: # try and parse the first 4 bytes as an ascii string
                s=''
                for i in range(4):
                        c = buf[i:i+1].decode('ASCII')
                        if c.isprintable():s+=c
                        else: break
                if s in FTYPE: return 0
        except (UnicodeDecodeError,Error):
                pass
        size = struct.unpack_from("<L",buf)[0]
        # unzipped size should be bigger than then len of the buffer
        # otherwise it isn't compressed
        if size < len(buf): return 0

        return size


def pack(buf):
        res=bytes([])
        while(len(buf)):
                nbytes=0x4000
                if nbytes > len(buf):
                        nbytes=len(buf)
                out = bytearray([0,0,0,0,0x78, 0x9C])
                adler32 = zlib.adler32(buf[:nbytes])
                try:
                        defl = zlib.compressobj(-1, wbits=-zlib.MAX_WBITS)
                        out+= defl.compress(buf[:nbytes])
                        out+= defl.flush()
                except Exception as ex:
                        raise Error("zlib exception")

                # chksum
                out+= struct.pack(">L", adler32)

                size=len(out)-4
                struct.pack_into('<L',out, 0, size|0x8000)
                pad = align(len(out),0x10) - len(out)
                out += bytes([0]*pad)
                res+=out
                buf=buf[nbytes:]
        return res

def unpack(buf, start=0):
        out=bytearray([])
        n=0 # for alignment
        while(len(buf)>6):
                size = struct.unpack_from('<L', buf)[0]
                size = size & ~0x8000
                if buf[4] == 0x78 and buf[5]==0x9C:
                        try:
                                out.extend( zlib.decompress(buf[6:],-zlib.MAX_WBITS))
                        except zlib.error as ex:
                                raise Error(ex)
                        end=align( n + size+4, 0x10)
                        buf[:end-n] = []
                        n=end
                else:
                        raise Error("invalid block @%#06x:%s"%(n,buf_str(buf[:10])))
        return out

def xor_buffer(buf, key):
        key = bytearray(key)
        key = (key*(len(XOR_PHRASE)//len(key)+1))[:len(XOR_PHRASE)]
        key[:]= [ (a^b) for a,b in zip(XOR_PHRASE,key) ]

        key = (key*( len(buf)//len(key)+1))[:len(buf)]
        buf[:] = [( (d^k) if (d and d!=k) else d)for d,k in zip(buf,key)]

def generate_key(num):
        A = num+0x3e7
        B = A*8 # FIXME: missing sth here to catch the possible overflow
        B -= A
        C = int(B / 0xB) + int(num % 0x11) + 0x1AC
        key=[]
        sh = 24
        while sh>=0:
                val = (C >> sh)&0xff
                if val > 0: key.append(val)
                sh-=8
        return bytes( key )

def replace_asset(exe, game_dir, asset, input_file):
        print("REPLACE %s in %s..."%(asset.alias,asset.lnk))
        print(" ..with %s"%input_file)
        stime=time()
        lnk_path=os.path.join(game_dir, asset.lnk_name)

        with open(lnk_path, 'rb') as fs:
                fs.seek(asset.lnk_ofs)
                dw = fs.read(4)

        with open(input_file, 'rb') as fs:
                data_in = fs.read()

#        if guess_is_compressed(dw):
        if asset.flags: #flags=0 => uncompressed
                final = struct.pack("<L", len(data_in))
                key = generate_key(len(data_in))
                s = time()
                packed = bytearray(pack(data_in))
                print(" ..zipped ( %.3f)"%(time()-s))
                s = time()
                xor_buffer(packed, key)
                print(" ..xored (%.3f)"%(time()-s))

                final += packed
                print(" ..using key:%s"%(buf_str(key)))
                print(" ..new data: %06x"%len(data_in))
                print(" ..old data: %06x"%asset.size)
                print(" ..new data(packed): %06x"%len(packed))
                print(" ..old data(packed): %06x"%asset.lnk_size)
        else:
                final = data_in
                print(" ..new data: %06x"%len(data_in))
                print(" ..old data: %6x"%asset.lnk_size)


        idx=asset.lnk_idx
        with open(lnk_path,'r+b') as fs:
                print(" ..correcting lnk directory @%#08x"%(28+idx*32))
                # note: has two size entries
                fs.seek(28+idx*32+12)
                fs.write( struct.pack("<L",len(final)))
                fs.seek(28+idx*32+20)
                fs.write( struct.pack("<L",len(final)))

                if asset.lnk_size >= len(final):
                        # can just overwrite it
                        ofs = asset.lnk_ofs
                else: # write to end of file
                        # FIXME: should actually rewrite the file from here
                        eof= os.stat(lnk_path).st_size

                        ofs=align( eof, 0x800)
                        fs.seek(16) #adjust file size
                        fs.write(struct.pack('<L',(ofs+len(final)) ))
                        print(" ..setting new size to:%#08x"%(ofs+len(final))) 

                        # write new offset
                        fs.seek(28+idx*32+4)
                        fs.write( struct.pack("<L",ofs))

                        # write padding
                        fs.seek(eof)
                        fs.write(bytes([0]*(ofs-eof)))

                asset.lnk_ofs=ofs
                print(" ..writing at ofs: %08x"%ofs)
                fs.seek(ofs)
                fs.write(final)

        fixup_exe(exe, asset, len(data_in))

        asset.size=len(data_in)
        asset.lnk_size=len(final)

        print(" ..timing: total:%.3f"%(time()-stime))
        print("REPLACE done")


def extract(game_dir, asset):
        stime=time()
        print("EXTRACT %s ..."%asset.alias)
        with open( os.path.join(game_dir, asset.lnk_name), 'rb') as fs:
                fs.seek(asset.lnk_ofs)
                buf = bytearray(asset.lnk_size)
                fs.readinto(buf)
        dt=time()
        print(" ..read idx:%d from ofs:%#08x size:%#06x"%(asset.lnk_idx,asset.lnk_ofs,asset.lnk_size))

        #size = guess_is_compressed(buf)
        size = struct.unpack_from("<L",buf)[0]

        # FIXME: verify this
        # ony when flags ==0, is the asset uncompressed
        if asset.flags: #size>0:
                size = struct.unpack_from("<L",buf)[0]
                buf[:4]=[] # discard
                key=generate_key(size)
                print(" ..key:0x%s"%buf_str(key))
                xor_buffer(buf, key)
                dt2=time()

                buf = unpack(buf)
                print(" ..unpacked %#06x/exp:%#06x"%(len(buf),asset.size))
                if len(buf)!=size:
                        print("error: ")
                dt3=time()
        else:
                print(" ..uncompressed")

        with open(asset.alias,'wb') as fs:
                fs.write(buf)
        end=time()
        if asset.flags:
                print(" ..timing: read:%.3f,xor:%.3f,unzip:%.3f write:%.3f total:%.3f"%(dt-stime,dt2-dt,dt3-dt2,end-dt3,end-stime))
        print("EXTRACT done => %s"%asset.alias)



def parse_lnk(game_path, lnk_name, idx_map, asset_map):
        with open(os.path.join(game_path, lnk_name),'rb') as fs:
                buf=fs.read(28)
                magic, u0, nfiles, u1, size, u2, u3 = struct.unpack("<4s6L",buf)
                # FIXME: check magic, but it seems there are different ones possble??
                #print("%s"%magic)
                buf=fs.read(32*nfiles)
                data=struct.unpack_from('<%dL'%(nfiles*8),buf)
                for idx,name in idx_map.items():
                        ofs, size = data[idx*8+1],data[idx*8+3]
                        asset = asset_map[name]
                        asset.lnk_idx = idx
                        asset.lnk_size=size
                        asset.lnk_ofs=ofs
                        asset.lnk_name=lnk_name


def parse_bin(game_path, fname, asset_map):
        lnk_id_map={}
        lnk_name_map={}

        #print("..parsing %s"%fname)
        with open( os.path.join(game_path,fname), 'rb') as fs:
                buf = fs.read()
        magic, nlnk,nfiles, lnk_ofs, fi_ofs, lnk_names_ofs, fi_names_ofs = struct.unpack_from("<4s6L", buf)
        if magic!=b'LFMO':
                raise Error("*err* invalid .bin file:%s"%path)

        # for each entry lnk_id, ofs of lnk_name
        data=struct.unpack_from("<%dL"%(nlnk*2),buf,lnk_ofs)
        for i in range(nlnk):
                lnk_name_map[ data[i*2] ] = get_zstr(buf,data[i*2+1])
                lnk_id_map [ data[i*2] ] = {}

        # for each entry: lnk_id, idx in lnkfile, ofs of name str
        data=struct.unpack_from("<%dL"%(nfiles*3),buf,fi_ofs)
        for i in range(nfiles):
                # note: skip initial '/'
                name = get_zstr(buf, data[i*3+2])[1:]
                lnk_id, lnk_idx,ofs = data[i*3:i*3+3]
                lnk_id_map[lnk_id][lnk_idx]=name

        # load lnk
        for k in lnk_name_map.keys():
                idx_map=lnk_id_map[k]
                lnk_name = lnk_name_map[k].replace('.order','.lnk')
                if len(idx_map.keys())==0:
                        continue
                parse_lnk(game_path, lnk_name, idx_map, asset_map)

def load_bin_data(game_path,asset_map):
        nfiles=0
        for entry in os.listdir(path=game_path):
                if entry.endswith('.bin'):
                        parse_bin(game_path, entry, asset_map)
                        nfiles+=1
        ndel=0
        for k in list(asset_map.keys()):
                if asset_map[k].lnk_ofs==None:
                        del asset_map[k]
                        ndel+=1
        print(" ..found %d assets in %d lnks/didn't find %d"%(len(asset_map.keys()),nfiles,ndel))



####################
# names.dat related stuff

# meh, probably want to rethink that whole thing
valid={ 'COS','HAIR','DLC','DLCU', 'FACE','C' }
def split_name(s):
        try:
                s=s[:s.rindex('.')]
                v=s.split('_')
                if v[1] in valid:
                        if len(v)>2: v[1]+=v[2]
                        return v[0],v[1]
                if v[1].rstrip('0123456789') in valid:
                        return v[0],v[1]
        except (ValueError,IndexError):
                pass
        raise Error

def build_cos_map(name_map, asset_map):
        cos_map={}
        bad=[]
        ncos=0
        ndec=0
        for k,v in name_map.items():
                if k not in asset_map:
                        bad.append(k)
                        continue
                asset_map[k].alias=v
                ndec+=1
                try:
                        char,cos =split_name(v)
                        ncos+=1
                        if char in cos_map:
                                if cos in cos_map[char]: cos_map[char][cos].append(k)
                                else: cos_map[char][cos]=[k]
                        else:
                                cos_map[char]={cos:[k]}
                except Error as e:
                        pass
        print(" ..found %d matching alias entries"%ndec)
        print(" ..found %d costume entries"%ncos)
        if len(bad):
                print(" ..found %d non-existing entries in names.dat"%len(bad))
                #for k in sorted(bad,key=lambda a: name_map[a]):
                 #       print("  %s\t%s"%(name_map[k],k))
        return cos_map


def load_names_data(path):
        name_map={}
        for entry in os.listdir(path=path):
                if not entry.endswith('.dat'): continue
                with open(os.path.join(path, entry),'r') as fs:
                        lines = fs.read().splitlines()
                        for i,line in enumerate(lines):
                                if not line or line[0]=='#': continue
                                try:
                                        p = line.split() # cut \r
                                        if len(p)==2:name_map[p[1]]=p[0]
                                        elif len(p)==3:name_map[p[1]]=p[2]
                                        else: raise ValueError
                                except ValueError:
                                        print("invalid data in %s at line:%d"%(entry,i+1))
        return name_map

class Filter:
        def __init__(self, func, val):
                self.func = func
                self.val=val

class LNKData:
        def __init__(self, exe_path):
                stime=time()
                print("LOADING...")
                self.exe_path = exe_path
                self.game_path = os.path.dirname(exe_path)
                self.name_map = load_names_data(DAT_DIR)
                self.asset_map = load_exe_info(self.exe_path)

                load_bin_data(self.game_path, self.asset_map)

                self.cos_map = build_cos_map(self.name_map, self.asset_map)

                print("LOADING finished(time:%.3f)"%(time()-stime))

                self.alias_map={}
                for pkg  in self.asset_map.values():
                        if pkg.name != pkg.alias: self.alias_map[pkg.alias]=pkg

                self.filters={}

        # changes -> list of tuple(pkg, path)
        def write_changes(self, changes):
                for pkg,path in sorted( changes, key=lambda it:it[0].lnk_name):
                        replace_asset(self.exe_path, self.game_path, pkg, path)

        def extract(self, asset): extract(self.game_path, asset)

        def add_filter(self,name, func, val=None):
                self.filters[name]=Filter(func,val)
        def remove_filter(self, name):
                if name in self.filters: del self.filters[name]
        def remove_filter_grp(self, match):
                sel = [ k for k in self.filters if match in k ]
                for f in sel: del self.filters[f]

        def get_filtered_list(self):
                assets=list(self.asset_map.values())
                for f in self.filters.values():
                        assets[:]=[ p for p in assets if f.func(p, f.val) ]
                return assets

        def set_names_filter(self,name, names):
                self.add_filter(name, lambda p,v: p.name in v, names)


#######################################
# gui stuff

class GUI(ttk.Frame):
        modes = ['costumes','modified','has alias', 'has no alias', 'everything']
        def __init__(self):
                ttk.Frame.__init__(self, padding="3 3 12 12")
                self.grid(sticky=(N,W,E,S))

                top=self.winfo_toplevel()
                top.resizable(False,False)
                self.master.title("%s %s"%(__appname__,__version__))

                self.data=None
                self.asset_list=[]
                self.work_queue={}

                w=self.make_exe_sel_widget()
                w.grid(column=0,row=0,sticky=(W,E),pady=4,padx=4)

                w=self.make_filter_widget()
                w.grid(column=0,row=1,sticky=(W,E),pady=4,padx=4)

                w=self.make_info_widget()
                w.grid(column=0,row=2,sticky=(W), pady=4,padx=4)

        def btn_sel_exe(self):
                path=filedialog.askopenfilename()
                self.load_data(path)

        def make_exe_sel_widget(self):
                frm = ttk.Frame(self)

                w=ttk.Label(frm, text="Game.exe:")
                w.grid(column=1,row=1)

                self.lbl_exe = StringVar()
                w=ttk.Label(frm, relief=SUNKEN,width=64,textvariable=self.lbl_exe)
                w.grid(column=2,row=1,padx=4)
                w=ttk.Button(frm,text='select', command=self.btn_sel_exe)
                w.grid(column=3,row=1,padx=4)
                return frm

        def make_info_widget(self):
                frm=ttk.Frame(self)
                self.lb = Listbox(frm, height=20,width=50)
                self.lb.bind('<<ListboxSelect>>', lambda x:self.lb_click())
                self.lb.grid(column=0,row=0,rowspan=5,sticky=(W),padx=5)
                self.val_nentries =StringVar()
                ttk.Label(frm,textvariable=self.val_nentries).grid(column=0,row=5,sticky=(W,N))

                ttk.Label(frm,text="Details:").grid(column=1,row=0,pady=1,sticky=(W))
                self.lb_d=Listbox(frm,height=14,width=40,relief=FLAT,state='disabled')
                self.lb_d.grid(column=1,columnspan=2,row=1,sticky=(N),padx=5)
                ttk.Button(frm,text="extract",width=16, command=self.extract).grid(column=1,row=3)
                ttk.Button(frm,text="extract all",width=16,command=self.extract_all).grid(column=2,row=3)
                ttk.Button(frm,text="swap",width=16,command=self.swap).grid(column=1,columnspan=2,row=4)
                ttk.Button(frm,text="write changes",width=16,command=self.write).grid(column=1,row=5)
                ttk.Button(frm,text="quit",width=16,command=self.quit).grid(column=2,row=5)

                return frm

        def make_filter_widget(self):
                frm=ttk.Frame(self)

                ttk.Label(frm,text="Filter Mode:").grid(column=0,row=0,padx=2)
                self.mode_val=StringVar()
                cb_mode=ttk.Combobox(frm,textvariable=self.mode_val,values=self.modes,state='readonly')
                cb_mode.grid(column=1,row=0,padx=2)
                cb_mode.bind('<<ComboboxSelected>>', lambda x: self.sel_mode())

                self.cur_char=StringVar()
                self.cb_chara = ttk.Combobox(frm,state='readonly',textvariable=self.cur_char)
                self.cb_chara.grid(column=2,row=0,padx=5)
                self.cb_chara.bind('<<ComboboxSelected>>',lambda x:self.sel_chara())

                self.cur_cos=StringVar()
                self.cb_cos = ttk.Combobox(frm,state='readonly',textvariable=self.cur_cos)
                self.cb_cos.grid(column=3,row=0,padx=5)
                self.cb_cos.bind('<<ComboboxSelected>>',lambda x:self.sel_cos())

                self.tx_filter = ttk.Entry(frm, width=32)
                self.tx_filter.grid(column=4,row=0,padx=5)
                self.tx_filter.bind('<KeyRelease>',lambda x:self.filter_keypress(self.tx_filter.get()))
                self.tx_filter.grid_remove()

                return frm


        def get_selected_asset(self):
                # why the hell does this return the line number as a string???
                sel = self.lb.curselection()
                if len(sel)==0: return None
                return self.asset_list[int(sel[0])]

        def lb_click(self):
                pkg = self.get_selected_asset()
                if not pkg: return
                lb=self.lb_d
                lb['state']='normal' #can't change content while disabled!?!
                lb.delete(0,END)
                lb.insert(END, "name:  %s"%pkg.name)
                lb.insert(END, "idx:  %#04x"%pkg.idx)
                lb.insert(END, "type:  %#02x"%pkg.type)
                lb.insert(END, "size:  %#06x (%.1f kb)"%(pkg.size,pkg.size/1024))
                lb.insert(END, "flags: %08x"%pkg.flags)
                fl=pkg.flags
                if fl:
                        s=" >"
                        if fl&(1<<31):s+=" ENC(1<<31)"
                        if fl&(1<<30):s+=" ??(1<<30)" # use size in crc_table??
                        if fl&(1<<29):s+=" CRC(1<<29)"
                        lb.insert(END, s)
                        if fl&0xffff:
                                lb.insert(END, " > crc idx: %#04x"%(fl&0xffff))
                        if fl&(1<<29):
                                lb.insert(END, "crc:  %#06x"%(pkg.crc))

                        if fl&(1<<31):
                                k=generate_key(pkg.size)
                                lb.insert(END,"enc key: 0x%s"%buf_str(k))

                lb.insert(END,"")
                lb.insert(END, "lnk:  %s"%pkg.lnk_name)
                lb.insert(END, "idx: %#06x"%pkg.lnk_idx)
                lb.insert(END, "ofs:  %#06x"%pkg.lnk_ofs)
                if pkg.lnk_size!=pkg.size:
                        lb.insert(END, "size(compressed): %#06x (%.1f kb)"%(pkg.lnk_size,pkg.lnk_size/1024))
                lb['state']='disabled'
                lb['disabledforeground']='#000000'

        def extract(self):
                asset = self.get_selected_asset()
                if asset: self.data.extract(asset)

        def extract_all(self):
                for asset in self.asset_list:
                        self.data.extract(asset)
        def write(self):
                todo = [ [self.data.asset_map[n],p] for n,p in self.work_queue.items() ]
                self.data.write_changes(todo)
                self.work_queue={}
                self.refresh_list()

        def swap(self):
                sel = self.lb.curselection()
                if len(sel)==0: return None
                pkg=self.asset_list[int(sel[0])]

                path=filedialog.askopenfilename()
                if path and os.path.exists(path):
                        self.work_queue[pkg.name]=path
                        self.refresh_list()

        def refresh_list(self):
                self.lb.delete(0,END)

                self.asset_list=self.data.get_filtered_list()
                self.asset_list.sort(key=lambda p: p.alias)
                for pkg in self.asset_list:
                        s = " "+pkg.alias
                        if pkg.name in self.work_queue:
                                s+=" => %s"%os.path.basename(self.work_queue[pkg.name])
                                self.lb.insert(END, s)
                                self.lb.itemconfig(END, foreground='#884400',selectforeground='#aa8800')
                        else:
                                self.lb.insert(END, s)
                self.val_nentries.set("%6d / %6d total assets"%(len(self.asset_list),len(self.data.asset_map)))

        def sel_chara(self):
                ch=self.cur_char.get()
                cos_map=self.data.cos_map[ch]
                self.cb_cos['values']=[ k for k in sorted(cos_map.keys())]
                self.cb_cos.current(0)
                self.sel_cos()
        def sel_cos(self):
                ch=self.cur_char.get()
                cos=self.cur_cos.get()
                x = self.data.cos_map[ch][cos]

                self.data.set_names_filter("m", x)
                self.refresh_list()

        def filter_by(self, key,val, idx):
                if key=='type':
                        try:val=int(val,0)
                        except ValueError: return 1
                        func=lambda p,v: p.type==v
                elif key=='idx':
                        try: val=int(val,0)
                        except ValueError: return 1
                        func=lambda p,v:p.idx==v
                elif key=='lnk': func=lambda p,v: v in p.lnk
                elif key=='flag':
                        try:val=int(val,0)
                        except ValueError: return 1
                        func=lambda p,v: (p.flags>>28)==v
                else: return 1
                if val: # only add if val has been type in
                        self.data.add_filter("custom%d"%idx,func,val)


        def filter_keypress(self,filter_string):
                self.data.remove_filter_grp("custom")
                idx=0
                err=0
                for f in filter_string.split():
                        if f[0]=='.':
                                v = f[1:].split(':')
                                if len(v)==2: err=self.filter_by(v[0],v[1],idx)
                        else:
                                self.data.add_filter("custom%d"%idx,lambda p,v: v in p.alias, f)
                        idx+=1

                if err: self.tx_filter.config(foreground='#CC0000')
                else: self.tx_filter.config(foreground='#000000')

                self.refresh_list()

        def sel_mode(self):
                mode = self.mode_val.get()
                if mode=='costumes':
                        self.tx_filter.grid_remove()
                        self.cb_chara.grid()
                        self.cb_cos.grid()
                        self.sel_chara()
                        return
                self.tx_filter.grid()
                self.cb_chara.grid_remove()
                self.cb_cos.grid_remove()

                if mode=='everything':
                        self.data.remove_filter("m")
                elif mode=='modified':
                        self.data.set_names_filter("m",self.work_queue.keys())
                elif mode=='has alias':
                        self.data.add_filter("m", lambda p,v: p.name!=p.alias )
                elif mode=='has no alias':
                        self.data.add_filter("m", lambda p,v: p.name==p.alias)
                self.refresh_list()

        def set_costume_filter(self, cos_map):
                charas= [ k for k in sorted(cos_map.keys()) ]
                self.cb_chara['values'] = charas
                self.cb_chara.current(0)
                self.sel_chara()

        def load_data(self,path):
                if not path and not os.path.exists(path):
                        # error
                        return

                self.lbl_exe.set(path)
                try:
                        self.data=LNKData(path)
                except Error as e:
                        pass #msg box

                self.set_costume_filter(self.data.cos_map)
                self.mode_val.set(self.modes[0])
                self.sel_mode()

        def run(self,exe_path=None):
                self.load_data(exe_path)
                self.master.mainloop()


def usage():
        print("Usage: %s [-exe]"%argv[0])
        print("\t-exe=path\tpath to DOA5LR's game.exe")


def run():
        opts={'exe':""}
        for arg in argv[1:]:
                if arg[0]!='-': continue
                try:
                        k,v = arg[1:].split('=')
                        if k in opts: opts[k]=v
                except ValueError:
                        print("invalid cmd line arguments:%s"%arg)
                        usage()
                        return
        gui = GUI()
        gui.run(opts['exe'])

if __name__ == '__main__':
        run()
